from __future__ import absolute_import, unicode_literals
import operator
import re
import sys
if sys.version_info >= (3, ):  # pragma: no cover
    # As in, Python 3
    from urllib.parse import urljoin
    STR_TYPE = str
else:  # Python 2
    from urlparse import urljoin
    STR_TYPE = basestring

from lxml import etree
from lxml.cssselect import CSSSelector
from django_inlinify.css_tools import CSSLoader, CSSParser

__all__ = ['Inlinify']


SINGLE_SELECTOR_REGEX = re.compile('^[\w\-.#]+$')
CDATA_REGEX = re.compile(r'<!\[CDATA\[(.*?)\]\]\>', re.DOTALL)


class Inlinify(object):

    def __init__(self,
                 css_files=None,
                 base_url=None,
                 preserve_internal_links=False,
                 preserve_inline_attachments=True,
                 method='html',
                 **kwargs):

        # attributes required by the URL parser
        self.base_url = base_url
        self.preserve_internal_links = preserve_internal_links
        self.preserve_inline_attachments = preserve_inline_attachments
        self.method = method

        if self.method not in ('html', 'xml'):
            raise ValueError('{} is not supported as a method'.format(method))

        # initialize parser and loader
        self.css_parser = CSSParser(**kwargs)
        self.css_source = CSSLoader(css_files)

    def transform(self, html, pretty_print=True, **kwargs):
        """Transform CSS into inline styles and inject them in the provided html
        """
        parser = None
        if self.method == 'html':
            parser = etree.HTMLParser()
        elif self.method == 'xml':
            parser = etree.XMLParser(ns_clean=False, resolve_entities=False)

        stripped = html.strip()
        tree = etree.fromstring(stripped, parser).getroottree()
        page = tree.getroot()

        # lxml inserts a doctype if none exists, so only include it in the root if it was in
        # the original html
        root = tree if stripped.startswith(tree.docinfo.doctype) else page

        assert page is not None

        # process style block
        rules = self._process_style_block(page)

        # process external style sheets
        rules.extend(self._process_external_files(page))

        # rules is a tuple of (specificity, selector, styles), where specificity is a tuple
        # ordered such that more specific rules sort larger.
        rules.sort(key=operator.itemgetter(0))

        original_styles = {}
        for __, selector, new_style in rules:

            # this check will avoid having to create a CSSSelector to then discover than there
            # is not any matching element in the document. Each selector can take up to 1.5ms, so
            # this is a huge time saver
            if SINGLE_SELECTOR_REGEX.match(selector) and selector not in stripped:
                continue

            sel = CSSSelector(selector)
            for item in sel(page):
                current_style = item.attrib.get('style', '')
                if item not in original_styles:
                    original_styles[item] = current_style
                self._update_element_style(item, current_style, new_style)

        # re-apply the original inline styles
        self._reapply_original_inline_styles(original_styles)

        # transform relative paths to absolute URLs if required
        self._transform_urls(page)

        # set some default options
        kwargs.setdefault('method', self.method)
        kwargs.setdefault('pretty_print', pretty_print)
        kwargs.setdefault('encoding', 'utf-8')

        html = etree.tostring(root, **kwargs).decode(kwargs['encoding'])

        # need to replace the "<![CDATA" style comments with "/*<![CDATA" comments to be valid XHTML
        if self.method == 'xml':
            html = CDATA_REGEX.sub(lambda m: '/*<![CDATA[*/%s/*]]>*/' % m.group(1), html)

        return html

    def _process_external_files(self, page):
        """Processes the provided external CSS files, if any
        """
        rules = []
        head = CSSSelector('head')(page)
        for index, css_body in enumerate(self.css_source):
            these_rules, these_leftover = self.css_parser.parse(css_body, index)
            rules.extend(these_rules)
            if these_leftover:
                style = etree.Element('style')
                style.attrib['type'] = 'text/css'
                if self.method == 'html':
                    style.text = these_leftover
                elif self.method == 'xml':
                    style.text = etree.CDATA(these_leftover)
                if head:
                    head[0].append(style)
        return rules

    def _process_style_block(self, page):
        """Processes the <style> block in the HTML
        """
        rules = []
        for index, element in enumerate(CSSSelector('style')(page)):
            # If we have a media attribute whose value is anything other than 'screen',
            # ignore the ruleset.
            media = element.attrib.get('media')
            if media and media != 'screen':
                continue

            css_body = element.text
            these_rules, these_leftover = self.css_parser.parse(css_body, index)
            rules.extend(these_rules)

        return rules

    def _reapply_original_inline_styles(self, original):
        """Re-applies all the initial inline styles
        """
        for item, inline_style in original.iteritems():
            if not inline_style:
                continue
            self._update_element_style(item, item.attrib.get('style', ''), inline_style)

    def _update_element_style(self, element, current, new):
        """Given an element, its current style and a new one, merges them and updates the element
        """
        new_style = self.css_parser.merge_styles(current, new)
        element.attrib['style'] = new_style
        self.css_parser.css_style_to_basic_html_attributes(element, new_style)

    def _transform_urls(self, page):
        if self.base_url:
            for attr in ('href', 'src'):
                for item in page.xpath("//@%s" % attr):
                    parent = item.getparent()
                    if (attr == 'href' and self.preserve_internal_links
                            and parent.attrib[attr].startswith('#')):
                        continue
                    if (attr == 'src' and self.preserve_inline_attachments
                            and parent.attrib[attr].startswith('cid:')):
                        continue
                    if not self.base_url.endswith('/'):
                        self.base_url += '/'
                    parent.attrib[attr] = urljoin(self.base_url, parent.attrib[attr].lstrip('/'))
        return page



# for testing, we'll use the default cache
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache'
    }
}

# django will complain if we don't include this
SECRET_KEY = 'dummysecret'



from setuptools import setup, find_packages

setup(
    name='django-inlinify',
    version='0.0.14',
    description="In-lines CSS into HTML and leverages Django's caching framework.",
    long_description="In-lines CSS into HTML and leverages Django's caching framework.",
    keywords='html lxml email mail style',
    author='Lyst Ltd.',
    author_email='devs@lyst.com',
    url='http://github.com/ssaw/django-inlinify/',
    license='Python',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(),
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['nose'],
    zip_safe=False,
    install_requires=[
        'lxml',
        'cssselect',
        'cssutils',
        'django>=1.5',
        'requests',
    ]
)



import re
import logging
import requests
import cssutils
from django.core.cache import get_cache, InvalidCacheBackendError
from django.conf import settings
from django_inlinify import defaults
from StringIO import StringIO
from contextlib import closing
from hashlib import md5

log = logging.getLogger('django_inlinify.css_loader')

DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME = getattr(
    settings,
    'DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME',
    defaults.DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME
)

# These pseudo selectors are ok to inline as they just filter the elements matched,
# as apposed to things like :hover or :focus which can't be inlined.
FILTER_PSEUDO_SELECTORS = [':last-child', ':first-child', ':nth-child']

# Regular expression to find number of different elements being targeted by a selector
ELEMENT_SELECTOR_REGEX = re.compile(r'(^|\s)\w')

# Regular expression to find all pseudo selectors in a selector
PSEUDO_SELECTOR_REGEX = re.compile(r':[a-z\-]+')


def load_cache(cache_name):
    """
    Tries to load the specified cache. If there is any problem, falls back to the default one

    Arguments:
        - str cache_name: the name of the cache backend to use

    Returns:
        cache object
    """
    if not cache_name:
        return get_cache(DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME)
    try:
        cache = get_cache(cache_name)
    except InvalidCacheBackendError:
        log.error('The cache you specified (%s) is not defined in settings. Falling back to '
                  'the default one (%s)', cache_name, DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME)
        cache = get_cache(DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME)
    return cache


class CSSLoader(object):
    """Class responsible for loading CSS files. Supports local and remote files
    """

    DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_PREFIX = getattr(settings,
                                         'DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_PREFIX',
                                         defaults.DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_PREFIX)
    DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_TTL = getattr(settings,
                                      'DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_TTL',
                                      defaults.DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_TTL)

    def __init__(self, files, cache_backend=None):
        self.files = files if files else []
        self.cache = load_cache(cache_backend)

    def _get_cache_key(self, filepath):
        return '%s_filecontents_%s_' % (self.DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_PREFIX, filepath)

    def _get_cached_contents(self, filename):
        return self.cache.get(self._get_cache_key(filename))

    def _get_file_contents_from_url(self, filepath):
        """Reads a remote file and returns its contents
        """
        response = requests.get(filepath, stream=True)
        if response.status_code != 200:
            raise ValueError('The CSS file you specified (%s) does not exist. Response (%s - %s)' %
                             (filepath, response.status_code, response.reason))

        with closing(StringIO()) as contents:
            for chunk in response.iter_content(512):
                contents.write(chunk)
        return contents.getvalue()

    def _get_file_contents_from_local_file(self, filepath):
        """Reads a file stored locally and returns its contents
        """
        with open(filepath) as f:
            contents = f.read()
        return contents

    def _read_file(self, filepath):
        """Reads the contents of the file located in the provided filepath

        Arguments:
            - str filepath: the path to the file

        Returns:
            the contents of the file
        """
        cached = self._get_cached_contents(filepath)
        if cached:
            return cached
        if filepath.startswith('http://') or filepath.startswith('https://'):
            contents = self._get_file_contents_from_url(filepath)
        else:
            contents = self._get_file_contents_from_local_file(filepath)

        self.cache.set(self._get_cache_key(filepath),
                       contents,
                       self.DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_TTL)

        return contents

    def __iter__(self):
        for f in self.files:
            yield self._read_file(f)


class CSSParser(object):
    """Class responsible for parsing CSS
    """

    DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_PREFIX = getattr(
        settings,
        'DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_PREFIX',
        defaults.DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_PREFIX
    )

    DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_TTL = getattr(
        settings,
        'DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_TTL',
        defaults.DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_TTL
    )

    DJANGO_INLINIFY_CSS_HTML_ATTRIBUTE_MAPPING = getattr(
        settings,
        'DJANGO_INLINIFY_CSS_HTML_ATTRIBUTE_MAPPING',
        defaults.DJANGO_INLINIFY_CSS_HTML_ATTRIBUTE_MAPPING
    )

    def __init__(self, cache_backend=None, **kwargs):
        self.cache = load_cache(cache_backend)
        self.include_star_selectors = kwargs.get('include_star_selectors', False)

    def _get_cache_key(self, css_body, index):
        h = md5(str(css_body)).hexdigest()
        return '%s_contents_%s_%s' % (self.DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_PREFIX, h, index)

    def _get_cached_css(self, css_body, index):
        return self.cache.get(self._get_cache_key(css_body, index))

    def parse(self, css_body, ruleset_index):
        """Extracts the rules from a CSS string. If they are cached, return those. Otherwise,
        extract them and cache them
        """
        cached = self._get_cached_css(css_body, ruleset_index)
        if cached:
            return cached
        parsed = self._parse_style_rules(css_body, ruleset_index)
        self.cache.set(self._get_cache_key(css_body, ruleset_index), parsed,
                       self.DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_TTL)
        return parsed

    def _parse_style_rules(self, css_body, ruleset_index):
        """Given a CSS string, extracts all its rules from it
        """
        leftover = []
        rules = []
        rule_index = 0

        # empty string
        if not css_body:
            return rules, leftover

        sheet = cssutils.parseString(css_body, validate=False)
        for rule in sheet:
            # handle media and font rules
            if rule.type in (rule.MEDIA_RULE, rule.FONT_FACE_RULE):
                leftover.append(rule)
                continue

            # only proceed for things we recognize
            if rule.type != rule.STYLE_RULE:
                continue

            bulk = ';'.join(
                u'{0}:{1}'.format(key, rule.style[key])
                for key in rule.style.keys()
            )
            selectors = (
                x.strip()
                for x in rule.selectorText.split(',')
                if x.strip() and not x.strip().startswith('@')
            )
            for selector in selectors:
                pseudos = [x.group(0) for x in PSEUDO_SELECTOR_REGEX.finditer(selector)]
                if any(pseudo not in FILTER_PSEUDO_SELECTORS for pseudo in pseudos):
                    leftover.append((selector, bulk))
                    continue
                elif '*' in selector and not self.include_star_selectors:
                    continue

                # Crudely calculate specificity
                id_count = selector.count('#')
                class_count = selector.count('.')
                element_count = len(ELEMENT_SELECTOR_REGEX.findall(selector))

                specificity = (id_count, class_count, element_count, ruleset_index, rule_index)

                rules.append((specificity, selector, bulk))
                rule_index += 1

        # we want to return a string, not those crazy CSSRule objects.
        # This will make serialization much faster
        leftover = self._css_rules_to_string(leftover)

        return rules, leftover

    def _make_important(self, bulk):
        """
        Marks every property in a string as `!important`
        """

        return ';'.join('%s !important' % p if not p.endswith('!important') else p for p in
                        bulk.split(';'))

    def _css_rules_to_string(self, rules):
        """
        Given a list of css rules returns a css string

        Arguments:
            - list rules: it can be either a list of cssutils.css.cssrule.CSSRule objects or tuples

        Returns:
            the CSS style string
        """
        lines = []
        for item in rules:
            if isinstance(item, tuple):
                k, v = item
                lines.append('%s {%s}' % (k, self._make_important(v)))
            elif item.type == item.FONT_FACE_RULE:
                lines.append(item.cssText)
            elif item.type == item.MEDIA_RULE:
                for rule in item.cssRules:
                    if isinstance(rule, cssutils.css.csscomment.CSSComment):
                        continue
                    for key in rule.style.keys():
                        rule.style[key] = (rule.style.getPropertyValue(key, False), '!important')
                lines.append(item.cssText)
        return '\n'.join(lines)

    def merge_styles(self, old_style, new_style):
        """
        Given two CSS styles, merges both so that any new style overrides the old ones

        Arguments:
            - str old_style: the old CSS style
            - str new_style: the new CSS style

        Returns:
            the new css style as a string
        """
        old_style_dict = self._css_string_to_dict(old_style)
        style_dict = self._css_string_to_dict(new_style)
        old_style_dict.update(style_dict)
        return '; '.join(['%s:%s' % (k, v) for k, v in sorted(old_style_dict.iteritems())])

    def _unbalanced(self, text):
        """
        Checks if there is an unbalanced parenthesis or braket in the provided text. Assumes that
        the text is processed left to right

        Arguments:
            - str text: the text to check

        Returns:
            true if its unbalanced, false otherwise
        """
        if text.count('(') and text.count('(') != text.count(')'):
            return True
        if text.count('{') and text.count('}') != text.count('}'):
            return True
        return False

    def _css_string_to_dict(self, css):
        """Given a string containing CSS, creates a dictionary out of it, where the keys are CSS
        properties and the values are their corresponding values

        This method assumes that CSS key-value pairs are separated by a semicolon. If this is not true,
        it can return unexpected results or even break

        Arguments:
            - str css: the css text

        Returns:
            a dictionary as described above
        """
        buff = ''
        css_properties = []
        for item in css.split(';'):
            # if we have any buffer, append the current item to the buffer
            if buff:
                item = buff + ';' + item
                buff = ''

            # if this breaks any parenthesis, brakets, buffer it and continue
            if self._unbalanced(item):
                buff = item
                continue

            # we are good to add the property
            if item.strip():
                css_properties.append(item.strip())

        # split every property into key, value and store them in a dict
        d = {}
        for css_property in css_properties:
            chunks = css_property.split(':', 1)
            d[chunks[0].strip()] = chunks[1].strip()
        return d

    def css_style_to_basic_html_attributes(self, element, style_content):
        """Given an element and styles like 'background-color:red; font-family:Arial' turn some of
        that into HTML attributes

        Note, the style_content can contain pseudoclasses like:
        '{color:red; border:1px solid green} :visited{border:1px solid green}'
        """
        if style_content.count('}') and style_content.count('{') == style_content.count('}'):
            style_content = style_content.split('}')[0][1:]

        mappings = self.DJANGO_INLINIFY_CSS_HTML_ATTRIBUTE_MAPPING

        for key, value in [
            x.split(':')
            for x in style_content.split(';') if len(x.split(':')) == 2
        ]:
            try:
                new_key, new_value = mappings.get(key.strip())
            except TypeError:
                continue
            else:
                element.attrib[new_key] = new_value(value)



# The default cache backend to use
DJANGO_INLINIFY_DEFAULT_CACHE_BACKEND_NAME = 'default'

# CSS Parser default settings
DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_PREFIX = 'django_inlinify_parsed_css_'
DJANGO_INLINIFY_CSSPARSER_CACHE_KEY_TTL = 60 * 60 * 24

DJANGO_INLINIFY_CSS_HTML_ATTRIBUTE_MAPPING = {
    'text-align': ('align', lambda value: value.strip()),
    'vertical-align': ('valign', lambda value: value.strip()),
    'background-color': ('bgcolor', lambda value: value.strip()),
    'width': ('width', lambda value: value.strip().replace('px', '')),
    'height': ('height', lambda value: value.strip().replace('px', '')),
    'cellspacing': ('cellspacing', lambda x: x.strip()),
    'cellpadding': ('cellpadding', lambda x: x.strip()),
}

# CSS Loader default settings
DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_PREFIX = 'django_inlinify_css_contents_'
DJANGO_INLINIFY_CSSLOADER_CACHE_KEY_TTL = 60 * 60 * 24
